# -*- coding: utf-8 -*-
# pylint: disable=chained-comparison,invalid-name,line-too-long,protected-access,too-many-arguments,too-many-boolean-expressions,too-many-branches,too-many-lines,too-many-locals,too-many-statements,unneeded-not,unused-import
u"""
Interfaces for the NAG Mark 29.2 `sort` Chapter.

``sort`` - Sorting and Searching

This module is concerned with sorting and searching numeric or character
data.
It handles only the simplest types of data structure and it is concerned
only with **internal** sorting and searching -- that is, sorting and
searching a set of data which can all be stored within the program.

If you have large files of data or complicated data structures to be
sorted or searched you should use a comprehensive sorting or searching
program or package.

For full information please refer to the NAG Library document

https://www.nag.com/numeric/nl/nagdoc_29.2/flhtml/m01/m01intro.html
"""

# NAG Copyright 2017-2023.

import collections as _collections

from ..base import sort as _b_sort
from ..base import utils as _b_utils
from ..library import _utils

def realvec_sort(rv, m1, order):
    u"""
    Sort a vector, real numbers.

    ``realvec_sort`` rearranges a vector of `float` numbers into
    ascending or descending order.

    For full information please refer to the NAG Library document for
    m01ca

    https://www.nag.com/numeric/nl/nagdoc_29.2/flhtml/m01/m01caf.html

    Parameters
    ----------
    rv : float, array-like, shape (m2)
        Elements `m1`-1 to m2-1 of `rv` must contain `float` values to
        be sorted.

    m1 : int
        Note: this argument represents an array index; the value you
        supply must be base-1 for compatibility with the NAG Engine.

        The index of the first element of `rv` to be sorted.

    order : str, length 1
        If `order` = 'A', the values will be sorted into ascending
        (i.e., nondecreasing) order.

        If `order` = 'D', into descending order.

    Returns
    -------
    rv : float, ndarray, shape (m2)
        These values are rearranged into sorted order.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `m1` = *<value>* and m2 = *<value>*.

            Constraint: `m1` <= m2.

        (`errno` 1)
            On entry, `m1` = *<value>*.

            Constraint: `m1` >= 1.

        (`errno` 1)
            On entry, m2 = *<value>*.

            Constraint: m2 >= 1.

        (`errno` 2)
            On entry, `order` has an illegal value: `order` = *<value>*.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    ``realvec_sort`` is based on Singleton's implementation of the
    'median-of-three' Quicksort algorithm (see Singleton (1969)), but
    with two additional modifications.
    First, small subfiles are sorted by an insertion sort on a separate
    final pass (see Sedgewick (1978)).
    Second, if a subfile is partitioned into two very unbalanced
    subfiles, the larger of them is flagged for special treatment:
    before it is partitioned, its end points are swapped with two random
    points within it; this makes the worst case behaviour extremely
    unlikely.

    References
    ----------
    Sedgewick, R, 1978, `Implementing Quicksort programs`, Comm. ACM
    (21), 847--857

    Singleton, R C, 1969, `An efficient algorithm for sorting with
    minimal storage: Algorithm 347`, Comm. ACM (12), 185--187
    """

    fun_name = 'naginterfaces.library.sort.realvec_sort'

    _rv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'rv'},
        rv,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m2_py = _b_utils._EngineFloat64ArrayType.get_shape(_rv, {'fun_name': fun_name, 'entity': 'rv'}, exp_rank=1)[0]
    _m2 = _b_utils._EngineIntScalarType(
        _m2_py, {'fun_name': fun_name, 'entity': '_m2'},
    ).to_ctypes()
    _b_sort.realvec_sort(
        _rv,
        m1,
        _m2,
        order,
    )
    _rv_py = _rv
    return _rv_py

def intvec_sort(iv, m1, order):
    u"""
    Sort a vector, integer numbers.

    ``intvec_sort`` rearranges a vector of integer numbers into
    ascending or descending order.

    For full information please refer to the NAG Library document for
    m01cb

    https://www.nag.com/numeric/nl/nagdoc_29.2/flhtml/m01/m01cbf.html

    Parameters
    ----------
    iv : int, array-like, shape (m2)
        Elements `m1`-1 to m2-1 of `iv` must contain integer values to
        be sorted.

    m1 : int
        Note: this argument represents an array index; the value you
        supply must be base-1 for compatibility with the NAG Engine.

        The index of the first element of `iv` to be sorted.

    order : str, length 1
        If `order` = 'A', the values will be sorted into ascending
        (i.e., nondecreasing) order.

        If `order` = 'D', into descending order.

    Returns
    -------
    iv : int, ndarray, shape (m2)
        These values are rearranged into sorted order.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `m1` = *<value>* and m2 = *<value>*.

            Constraint: `m1` <= m2.

        (`errno` 1)
            On entry, `m1` = *<value>*.

            Constraint: `m1` >= 1.

        (`errno` 1)
            On entry, m2 = *<value>*.

            Constraint: m2 >= 1.

        (`errno` 2)
            On entry, `order` has an illegal value: `order` = *<value>*.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    ``intvec_sort`` is based on Singleton's implementation of the
    'median-of-three' Quicksort algorithm (see Singleton (1969)), but
    with two additional modifications.
    First, small subfiles are sorted by an insertion sort on a separate
    final pass (see Sedgewick (1978)) Second, if a subfile is
    partitioned into two very unbalanced subfiles, the larger of them is
    flagged for special treatment: before it is partitioned, its end
    points are swapped with two random points within it; this makes the
    worst case behaviour extremely unlikely.

    References
    ----------
    Sedgewick, R, 1978, `Implementing Quicksort programs`, Comm. ACM
    (21), 847--857

    Singleton, R C, 1969, `An efficient algorithm for sorting with
    minimal storage: Algorithm 347`, Comm. ACM (12), 185--187
    """

    fun_name = 'naginterfaces.library.sort.intvec_sort'

    _iv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'iv'},
        iv,
        copy=True,
    )
    _m2_py = _b_utils._EngineIntArrayType.get_shape(_iv, {'fun_name': fun_name, 'entity': 'iv'}, exp_rank=1)[0]
    _m2 = _b_utils._EngineIntScalarType(
        _m2_py, {'fun_name': fun_name, 'entity': '_m2'},
    ).to_ctypes()
    _b_sort.intvec_sort(
        _iv,
        m1,
        _m2,
        order,
    )
    _iv_py = _iv
    return _iv_py

def charvec_sort(ch, m1, l1, l2, order):
    u"""
    Sort a vector, character data.

    ``charvec_sort`` rearranges a vector of character data so that a
    specified substring is in ASCII or reverse ASCII order.

    For full information please refer to the NAG Library document for
    m01cc

    https://www.nag.com/numeric/nl/nagdoc_29.2/flhtml/m01/m01ccf.html

    Parameters
    ----------
    ch : str, array-like, shape (m2)
        Elements `m1`-1 to m2-1 of `ch` must contain character data to
        be sorted.

    m1 : int
        Note: this argument represents an array index; the value you
        supply must be base-1 for compatibility with the NAG Engine.

        The index of the first element of `ch` to be sorted.

    l1 : int
        Only the substring (`l1`:`l2`) of each element of `ch` is to be
        used in determining the sorted order.

    l2 : int
        Only the substring (`l1`:`l2`) of each element of `ch` is to be
        used in determining the sorted order.

    order : str, length 1
        If `order` = 'A', the values will be sorted into ASCII order.

        If `order` = 'R', into reverse ASCII order.

    Returns
    -------
    ch : str, ndarray, shape (m2)
        These values are rearranged into sorted order.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `l2` = *<value>* and len(`ch`[0]) = *<value>*.

            Constraint: `l2` <= len(`ch`[0]).

        (`errno` 1)
            On entry, `l1` = *<value>* and `l2` = *<value>*.

            Constraint: `l1` <= `l2`.

        (`errno` 1)
            On entry, `l1` = *<value>*.

            Constraint: `l1` >= 1.

        (`errno` 1)
            On entry, `l2` = *<value>*.

            Constraint: `l2` >= 1.

        (`errno` 1)
            On entry, `m1` = *<value>* and m2 = *<value>*.

            Constraint: `m1` <= m2.

        (`errno` 1)
            On entry, `m1` = *<value>*.

            Constraint: `m1` >= 1.

        (`errno` 1)
            On entry, m2 = *<value>*.

            Constraint: m2 >= 1.

        (`errno` 2)
            On entry, `order` has an illegal value: `order` = *<value>*.

        (`errno` 3)
            On entry, len(`ch`[0]) = *<value>*.

            Constraint: len(`ch`[0]) <= 255.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    ``charvec_sort`` is based on Singleton's implementation of the
    'median-of-three' Quicksort algorithm (see Singleton (1969)), but
    with two additional modifications.
    First, small subfiles are sorted by an insertion sort on a separate
    final pass (see Sedgewick (1978)) Second, if a subfile is
    partitioned into two very unbalanced subfiles, the larger of them is
    flagged for special treatment: before it is partitioned, its end
    points are swapped with two random points within it; this makes the
    worst case behaviour extremely unlikely.

    Only the substring (`l1`:`l2`) of each element of the array `ch` is
    used to determine the sorted order, but the entire elements are
    rearranged into sorted order.

    References
    ----------
    Sedgewick, R, 1978, `Implementing Quicksort programs`, Comm. ACM
    (21), 847--857

    Singleton, R C, 1969, `An efficient algorithm for sorting with
    minimal storage: Algorithm 347`, Comm. ACM (12), 185--187
    """

    fun_name = 'naginterfaces.library.sort.charvec_sort'

    _ch = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ch'},
        ch,
        copy=True,
    )
    _m2_py = _b_utils._EngineCharArrayType.get_shape(_ch, {'fun_name': fun_name, 'entity': 'ch'}, exp_rank=1)[0]
    _m2 = _b_utils._EngineIntScalarType(
        _m2_py, {'fun_name': fun_name, 'entity': '_m2'},
    ).to_ctypes()
    _b_sort.charvec_sort(
        _ch,
        m1,
        _m2,
        l1,
        l2,
        order,
    )
    _ch_py = _ch
    return _ch_py

def realvec_rank(rv, m1, order):
    u"""
    Rank a vector, real numbers.

    ``realvec_rank`` ranks a vector of `float` numbers in ascending or
    descending order.

    For full information please refer to the NAG Library document for
    m01da

    https://www.nag.com/numeric/nl/nagdoc_29.2/flhtml/m01/m01daf.html

    Parameters
    ----------
    rv : float, array-like, shape (m2)
        Elements `m1`-1 to m2-1 of `rv` must contain `float` values to
        be ranked.

    m1 : int
        Note: this argument represents an array index; the value you
        supply must be base-1 for compatibility with the NAG Engine.

        The index of the first element of `rv` to be ranked.

    order : str, length 1
        If `order` = 'A', the values will be ranked in ascending (i.e.,
        nondecreasing) order.

        If `order` = 'D', into descending order.

    Returns
    -------
    irank : int, ndarray, shape (m2)
        Elements `m1`-1 to m2-1 of `irank` contain the ranks of the
        corresponding elements of `rv`. Note that the ranks are in the
        range `m1` to m2: thus, if `rv`[i-1] is the first element in the
        rank order, `irank`[i-1] is set to `m1`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `m1` = *<value>* and m2 = *<value>*.

            Constraint: `m1` <= m2.

        (`errno` 1)
            On entry, `m1` = *<value>*.

            Constraint: `m1` >= 1.

        (`errno` 1)
            On entry, m2 = *<value>*.

            Constraint: m2 >= 1.

        (`errno` 2)
            On entry, `order` has an illegal value: `order` = *<value>*.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    ``realvec_rank`` uses a variant of list-merging, as described on
    pages 165--166 in Knuth (1973).
    The function takes advantage of natural ordering in the data, and
    uses a simple list insertion in a preparatory pass to generate
    ordered lists of length at least 10.
    The ranking is stable: equal elements preserve their ordering in the
    input data.

    References
    ----------
    Knuth, D E, 1973, `The Art of Computer Programming (Volume 3)`, (2nd
    Edition), Addison--Wesley
    """

    fun_name = 'naginterfaces.library.sort.realvec_rank'

    _rv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'rv'},
        rv,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m2_py = _b_utils._EngineFloat64ArrayType.get_shape(_rv, {'fun_name': fun_name, 'entity': 'rv'}, exp_rank=1)[0]
    _m2 = _b_utils._EngineIntScalarType(
        _m2_py, {'fun_name': fun_name, 'entity': '_m2'},
    ).to_ctypes()
    _irank = _b_utils._EngineIntArrayType.empty_ndarray(
        (_m2_py,),
    )
    _b_sort.realvec_rank(
        _rv,
        m1,
        _m2,
        order,
        _irank,
    )
    _irank_py = _irank
    return _irank_py

def intvec_rank(iv, m1, order):
    u"""
    Rank a vector, integer numbers.

    ``intvec_rank`` ranks a vector of integer numbers in ascending or
    descending order.

    For full information please refer to the NAG Library document for
    m01db

    https://www.nag.com/numeric/nl/nagdoc_29.2/flhtml/m01/m01dbf.html

    Parameters
    ----------
    iv : int, array-like, shape (m2)
        Elements `m1`-1 to m2-1 of `iv` must contain integer values to
        be ranked.

    m1 : int
        Note: this argument represents an array index; the value you
        supply must be base-1 for compatibility with the NAG Engine.

        The index of the first element of `iv` to be ranked.

    order : str, length 1
        If `order` = 'A', the values will be ranked in ascending (i.e.,
        nondecreasing) order.

        If `order` = 'D', into descending order.

    Returns
    -------
    irank : int, ndarray, shape (m2)
        Elements `m1`-1 to m2-1 of `irank` contain the ranks of the
        corresponding elements of `iv`. Note that the ranks are in the
        range `m1` to m2: thus, if `iv`[i-1] is the first element in the
        rank order, `irank`[i-1] is set to `m1`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `m1` = *<value>* and m2 = *<value>*.

            Constraint: `m1` <= m2.

        (`errno` 1)
            On entry, `m1` = *<value>*.

            Constraint: `m1` >= 1.

        (`errno` 1)
            On entry, m2 = *<value>*.

            Constraint: m2 >= 1.

        (`errno` 2)
            On entry, `order` has an illegal value: `order` = *<value>*.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    ``intvec_rank`` uses a variant of list-merging, as described on
    pages 165--166 in Knuth (1973).
    The function takes advantage of natural ordering in the data, and
    uses a simple list insertion in a preparatory pass to generate
    ordered lists of length at least 10.
    The ranking is stable: equal elements preserve their ordering in the
    input data.

    References
    ----------
    Knuth, D E, 1973, `The Art of Computer Programming (Volume 3)`, (2nd
    Edition), Addison--Wesley
    """

    fun_name = 'naginterfaces.library.sort.intvec_rank'

    _iv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'iv'},
        iv,
        copy=False,
    )
    _m2_py = _b_utils._EngineIntArrayType.get_shape(_iv, {'fun_name': fun_name, 'entity': 'iv'}, exp_rank=1)[0]
    _m2 = _b_utils._EngineIntScalarType(
        _m2_py, {'fun_name': fun_name, 'entity': '_m2'},
    ).to_ctypes()
    _irank = _b_utils._EngineIntArrayType.empty_ndarray(
        (_m2_py,),
    )
    _b_sort.intvec_rank(
        _iv,
        m1,
        _m2,
        order,
        _irank,
    )
    _irank_py = _irank
    return _irank_py

def charvec_rank(ch, m1, l1, l2, order):
    u"""
    Rank a vector, character data.

    ``charvec_rank`` ranks a vector of character data in ASCII or
    reverse ASCII order of a specified substring.

    For full information please refer to the NAG Library document for
    m01dc

    https://www.nag.com/numeric/nl/nagdoc_29.2/flhtml/m01/m01dcf.html

    Parameters
    ----------
    ch : str, array-like, shape (m2)
        Elements `m1`-1 to m2-1 of `ch` must contain character data to
        be ranked.

    m1 : int
        Note: this argument represents an array index; the value you
        supply must be base-1 for compatibility with the NAG Engine.

        The index of the first element of `ch` to be ranked.

    l1 : int
        Only the substring (`l1`:`l2`) of each element of `ch` is to be
        used in determining the rank order.

    l2 : int
        Only the substring (`l1`:`l2`) of each element of `ch` is to be
        used in determining the rank order.

    order : str, length 1
        If `order` = 'A', the values will be ranked in ASCII order.

        If `order` = 'R', in reverse ASCII order.

    Returns
    -------
    irank : int, ndarray, shape (m2)
        Elements `m1`-1 to m2-1 of `irank` contain the ranks of the
        corresponding elements of `ch`. Note that the ranks are in the
        range `m1` to m2: thus, if `ch`[i-1] is the first element in the
        rank order, `irank`[i-1] is set to `m1`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `l2` = *<value>* and len(`ch`[0]) = *<value>*.

            Constraint: `l2` <= len(`ch`[0]).

        (`errno` 1)
            On entry, `l1` = *<value>* and `l2` = *<value>*.

            Constraint: `l1` <= `l2`.

        (`errno` 1)
            On entry, `l1` = *<value>*.

            Constraint: `l1` >= 1.

        (`errno` 1)
            On entry, `l2` = *<value>*.

            Constraint: `l2` >= 1.

        (`errno` 1)
            On entry, `m1` = *<value>* and m2 = *<value>*.

            Constraint: `m1` <= m2.

        (`errno` 1)
            On entry, `m1` = *<value>*.

            Constraint: `m1` >= 1.

        (`errno` 1)
            On entry, m2 = *<value>*.

            Constraint: m2 >= 1.

        (`errno` 2)
            On entry, `order` has an illegal value: `order` = *<value>*.

        (`errno` 3)
            On entry, len(`ch`[0]) = *<value>*.

            Constraint: len(`ch`[0]) <= 255.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    ``charvec_rank`` uses a variant of list-merging, as described on
    pages 165--166 in Knuth (1973).
    The function takes advantage of natural ordering in the data, and
    uses a simple list insertion in a preparatory pass to generate
    ordered lists of length at least 10.
    The ranking is stable: equal elements preserve their ordering in the
    input data.

    Only the substring (`l1`:`l2`) of each element of the array `ch` is
    used to determine the rank order.

    References
    ----------
    Knuth, D E, 1973, `The Art of Computer Programming (Volume 3)`, (2nd
    Edition), Addison--Wesley
    """

    fun_name = 'naginterfaces.library.sort.charvec_rank'

    _ch = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ch'},
        ch,
        copy=False,
    )
    _m2_py = _b_utils._EngineCharArrayType.get_shape(_ch, {'fun_name': fun_name, 'entity': 'ch'}, exp_rank=1)[0]
    _m2 = _b_utils._EngineIntScalarType(
        _m2_py, {'fun_name': fun_name, 'entity': '_m2'},
    ).to_ctypes()
    _irank = _b_utils._EngineIntArrayType.empty_ndarray(
        (_m2_py,),
    )
    _b_sort.charvec_rank(
        _ch,
        m1,
        _m2,
        l1,
        l2,
        order,
        _irank,
    )
    _irank_py = _irank
    return _irank_py

def realmat_rank_rows(rm, m1, n1, order):
    u"""
    Rank rows of a matrix, real numbers.

    ``realmat_rank_rows`` ranks the rows of a matrix of `float` numbers
    in ascending or descending order.

    For full information please refer to the NAG Library document for
    m01de

    https://www.nag.com/numeric/nl/nagdoc_29.2/flhtml/m01/m01def.html

    Parameters
    ----------
    rm : float, array-like, shape (m2, n2)
        Columns `n1` to n2 of rows `m1`-1 to m2-1 of `rm` must contain
        `float` data to be ranked.

    m1 : int
        Note: this argument represents an array index; the value you
        supply must be base-1 for compatibility with the NAG Engine.

        The index of the first row of `rm` to be ranked.

    n1 : int
        The index of the first column of `rm` to be used.

    order : str, length 1
        If `order` = 'A', the rows will be ranked in ascending (i.e.,
        nondecreasing) order.

        If `order` = 'D', into descending order.

    Returns
    -------
    irank : int, ndarray, shape (m2)
        Elements `m1`-1 to m2-1 of `irank` contain the ranks of the
        corresponding rows of `rm`. Note that the ranks are in the range
        `m1` to m2: thus, if the ith row of `rm` is the first in the
        rank order, `irank`[i-1] is set to `m1`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, m2 = *<value>* and ldm = *<value>*.

            Constraint: m2 <= ldm.

        (`errno` 1)
            On entry, `n1` = *<value>* and n2 = *<value>*.

            Constraint: `n1` <= n2.

        (`errno` 1)
            On entry, `n1` = *<value>*.

            Constraint: `n1` >= 1.

        (`errno` 1)
            On entry, n2 = *<value>*.

            Constraint: n2 >= 1.

        (`errno` 1)
            On entry, `m1` = *<value>* and m2 = *<value>*.

            Constratint: `m1` <= m2.

        (`errno` 1)
            On entry, `m1` = *<value>*.

            Constraint: `m1` >= 1.

        (`errno` 1)
            On entry, m2 = *<value>*.

            Constraint: m2 >= 1.

        (`errno` 2)
            On entry, `order` has an illegal value: `order` = *<value>*.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    ``realmat_rank_rows`` ranks rows `m1` to m2 of a matrix, using the
    data in columns `n1` to n2 of those rows.
    The ordering is determined by first ranking the data in column `n1`,
    then ranking any tied rows according to the data in column `n1`+1,
    and so on up to column n2.

    ``realmat_rank_rows`` uses a variant of list-merging, as described
    on pages 165--166 in Knuth (1973).
    The function takes advantage of natural ordering in the data, and
    uses a simple list insertion in a preparatory pass to generate
    ordered lists of length at least 10.
    The ranking is stable: equal rows preserve their ordering in the
    input data.

    References
    ----------
    Knuth, D E, 1973, `The Art of Computer Programming (Volume 3)`, (2nd
    Edition), Addison--Wesley
    """

    fun_name = 'naginterfaces.library.sort.realmat_rank_rows'

    _rm = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'rm'},
        rm,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m2_py = _b_utils._EngineFloat64ArrayType.get_shape(_rm, {'fun_name': fun_name, 'entity': 'rm'}, exp_rank=2)[0]
    _m2 = _b_utils._EngineIntScalarType(
        _m2_py, {'fun_name': fun_name, 'entity': '_m2'},
    ).to_ctypes()
    _n2_py = _b_utils._EngineFloat64ArrayType.get_shape(_rm, {'fun_name': fun_name, 'entity': 'rm'}, exp_rank=2)[1]
    _n2 = _b_utils._EngineIntScalarType(
        _n2_py, {'fun_name': fun_name, 'entity': '_n2'},
    ).to_ctypes()
    _irank = _b_utils._EngineIntArrayType.empty_ndarray(
        (_m2_py,),
    )
    _b_sort.realmat_rank_rows(
        _rm,
        m1,
        _m2,
        n1,
        _n2,
        order,
        _irank,
    )
    _irank_py = _irank
    return _irank_py

def intmat_rank_rows(im, m1, n1, order):
    u"""
    Rank rows of a matrix, integer numbers.

    ``intmat_rank_rows`` ranks the rows of a matrix of integer numbers
    in ascending or descending order.

    For full information please refer to the NAG Library document for
    m01df

    https://www.nag.com/numeric/nl/nagdoc_29.2/flhtml/m01/m01dff.html

    Parameters
    ----------
    im : int, array-like, shape (m2, n2)
        Columns `n1` to n2 of rows `m1`-1 to m2-1 of `im` must contain
        `int` data to be ranked.

    m1 : int
        Note: this argument represents an array index; the value you
        supply must be base-1 for compatibility with the NAG Engine.

        The index of the first row of `im` to be ranked.

    n1 : int
        The index of the first column of `im` to be used.

    order : str, length 1
        If `order` = 'A', the rows will be ranked in ascending (i.e.,
        nondecreasing) order.

        If `order` = 'D', into descending order.

    Returns
    -------
    irank : int, ndarray, shape (m2)
        Elements `m1`-1 to m2-1 of `irank` contain the ranks of the
        corresponding rows of `im`. Note that the ranks are in the range
        `m1` to m2: thus, if the ith row of `im` is the first in the
        rank order, `irank`[i-1] is set to `m1`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, m2 = *<value>* and ldm = *<value>*.

            Constraint: m2 <= ldm.

        (`errno` 1)
            On entry, `n1` = *<value>* and n2 = *<value>*.

            Constraint: `n1` <= n2.

        (`errno` 1)
            On entry, `n1` = *<value>*.

            Constraint: `n1` >= 1.

        (`errno` 1)
            On entry, n2 = *<value>*.

            Constraint: n2 >= 1.

        (`errno` 1)
            On entry, `m1` = *<value>* and m2 = *<value>*.

            Constraint: `m1` <= m2.

        (`errno` 1)
            On entry, `m1` = *<value>*.

            Constraint: `m1` >= 1.

        (`errno` 1)
            On entry, m2 = *<value>*.

            Constraint: m2 >= 1.

        (`errno` 2)
            On entry, `order` has an illegal value: `order` = *<value>*.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    ``intmat_rank_rows`` ranks rows `m1` to m2 of a matrix, using the
    data in columns `n1` to n2 of those rows.
    The ordering is determined by first ranking the data in column `n1`,
    then ranking any tied rows according to the data in column `n1`+1,
    and so on up to column n2.

    ``intmat_rank_rows`` uses a variant of list-merging, as described on
    pages 165--166 in Knuth (1973).
    The function takes advantage of natural ordering in the data, and
    uses a simple list insertion in a preparatory pass to generate
    ordered lists of length at least 10.
    The ranking is stable: equal rows preserve their ordering in the
    input data.

    References
    ----------
    Knuth, D E, 1973, `The Art of Computer Programming (Volume 3)`, (2nd
    Edition), Addison--Wesley
    """

    fun_name = 'naginterfaces.library.sort.intmat_rank_rows'

    _im = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'im'},
        im,
        copy=False,
    )
    _m2_py = _b_utils._EngineIntArrayType.get_shape(_im, {'fun_name': fun_name, 'entity': 'im'}, exp_rank=2)[0]
    _m2 = _b_utils._EngineIntScalarType(
        _m2_py, {'fun_name': fun_name, 'entity': '_m2'},
    ).to_ctypes()
    _n2_py = _b_utils._EngineIntArrayType.get_shape(_im, {'fun_name': fun_name, 'entity': 'im'}, exp_rank=2)[1]
    _n2 = _b_utils._EngineIntScalarType(
        _n2_py, {'fun_name': fun_name, 'entity': '_n2'},
    ).to_ctypes()
    _irank = _b_utils._EngineIntArrayType.empty_ndarray(
        (_m2_py,),
    )
    _b_sort.intmat_rank_rows(
        _im,
        m1,
        _m2,
        n1,
        _n2,
        order,
        _irank,
    )
    _irank_py = _irank
    return _irank_py

def realmat_rank_columns(rm, m1, n1, order):
    u"""
    Rank columns of a matrix, real numbers.

    ``realmat_rank_columns`` ranks the columns of a matrix of `float`
    numbers in ascending or descending order.

    For full information please refer to the NAG Library document for
    m01dj

    https://www.nag.com/numeric/nl/nagdoc_29.2/flhtml/m01/m01djf.html

    Parameters
    ----------
    rm : float, array-like, shape (m2, n2)
        Rows `m1`-1 to m2-1 of columns `n1` to n2 of `rm` must contain
        `float` data to be ranked.

    m1 : int
        Note: this argument represents an array index; the value you
        supply must be base-1 for compatibility with the NAG Engine.

        The index of the first row of `rm` to be used.

    n1 : int
        The index of the first column of `rm` to be ranked.

    order : str, length 1
        If `order` = 'A', the columns will be ranked in ascending (i.e.,
        nondecreasing) order.

        If `order` = 'D', into descending order.

    Returns
    -------
    irank : int, ndarray, shape (n2)
        Elements `n1` to n2 of `irank` contain the ranks of the
        corresponding columns of `rm`. Note that the ranks are in the
        range `n1` to n2: thus, if the ith column of `rm` is the first
        in the rank order, `irank`[i-1] is set to `n1`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, m2 = *<value>* and ldm = *<value>*.

            Constraint: m2 <= ldm.

        (`errno` 1)
            On entry, `n1` = *<value>* and n2 = *<value>*.

            Constraint: `n1` <= n2.

        (`errno` 1)
            On entry, `n1` = *<value>*.

            Constraint: `n1` >= 1.

        (`errno` 1)
            On entry, n2 = *<value>*.

            Constraint: n2 >= 1.

        (`errno` 1)
            On entry, `m1` = *<value>* and m2 = *<value>*.

            Constraint: `m1` <= m2.

        (`errno` 1)
            On entry, `m1` = *<value>*.

            Constraint: `m1` >= 1.

        (`errno` 1)
            On entry, m2 = *<value>*.

            Constraint: m2 >= 1.

        (`errno` 2)
            On entry, `order` has an illegal value: `order` = *<value>*.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    ``realmat_rank_columns`` ranks columns `n1` to n2 of a matrix, using
    the data in rows `m1` to m2 of those columns.
    The ordering is determined by first ranking the data in row `m1`,
    then ranking any tied columns according to the data in row `m1`+1,
    and so on up to row m2.

    ``realmat_rank_columns`` uses a variant of list-merging, as
    described on pages 165--166 in Knuth (1973).
    The function takes advantage of natural ordering in the data, and
    uses a simple list insertion in a preparatory pass to generate
    ordered lists of length at least 10.
    The ranking is stable: equal columns preserve their ordering in the
    input data.

    References
    ----------
    Knuth, D E, 1973, `The Art of Computer Programming (Volume 3)`, (2nd
    Edition), Addison--Wesley
    """

    fun_name = 'naginterfaces.library.sort.realmat_rank_columns'

    _rm = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'rm'},
        rm,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m2_py = _b_utils._EngineFloat64ArrayType.get_shape(_rm, {'fun_name': fun_name, 'entity': 'rm'}, exp_rank=2)[0]
    _m2 = _b_utils._EngineIntScalarType(
        _m2_py, {'fun_name': fun_name, 'entity': '_m2'},
    ).to_ctypes()
    _n2_py = _b_utils._EngineFloat64ArrayType.get_shape(_rm, {'fun_name': fun_name, 'entity': 'rm'}, exp_rank=2)[1]
    _n2 = _b_utils._EngineIntScalarType(
        _n2_py, {'fun_name': fun_name, 'entity': '_n2'},
    ).to_ctypes()
    _irank = _b_utils._EngineIntArrayType.empty_ndarray(
        (_n2_py,),
    )
    _b_sort.realmat_rank_columns(
        _rm,
        m1,
        _m2,
        n1,
        _n2,
        order,
        _irank,
    )
    _irank_py = _irank
    return _irank_py

def intmat_rank_columns(im, m1, n1, order):
    u"""
    Rank columns of a matrix, integer numbers.

    ``intmat_rank_columns`` ranks the columns of a matrix of integer
    numbers in ascending or descending order.

    For full information please refer to the NAG Library document for
    m01dk

    https://www.nag.com/numeric/nl/nagdoc_29.2/flhtml/m01/m01dkf.html

    Parameters
    ----------
    im : int, array-like, shape (m2, n2)
        Rows `m1`-1 to m2-1 of columns `n1` to n2 of `im` must contain
        integer data to be ranked.

    m1 : int
        Note: this argument represents an array index; the value you
        supply must be base-1 for compatibility with the NAG Engine.

        The index of the first row of `im` to be used.

    n1 : int
        The index of the first column of `im` to be ranked.

    order : str, length 1
        If `order` = 'A', the columns will be ranked in ascending (i.e.,
        nondecreasing) order.

        If `order` = 'D', into descending order.

    Returns
    -------
    irank : int, ndarray, shape (n2)
        Elements `n1` to n2 of `irank` contain the ranks of the
        corresponding columns of `im`. Note that the ranks are in the
        range `n1` to n2: thus, if the ith column of `im` is the first
        in the rank order, `irank`[i-1] is set to `n1`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, m2 = *<value>* and ldm = *<value>*.

            Constraint: m2 <= ldm.

        (`errno` 1)
            On entry, `n1` = *<value>* and n2 = *<value>*.

            Constraint: `n1` <= n2.

        (`errno` 1)
            On entry, `n1` = *<value>*.

            Constraint: `n1` >= 1.

        (`errno` 1)
            On entry, n2 = *<value>*.

            Constraint: n2 >= 1.

        (`errno` 1)
            On entry, `m1` = *<value>* and m2 = *<value>*.

            Constraint: `m1` <= m2.

        (`errno` 1)
            On entry, `m1` = *<value>*.

            Constraint: `m1` >= 1.

        (`errno` 1)
            On entry, m2 = *<value>*.

            Constraint: m2 >= 1.

        (`errno` 2)
            On entry, `order` has an illegal value: `order` = *<value>*.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    ``intmat_rank_columns`` ranks columns `n1` to n2 of a matrix, using
    the data in rows `m1` to m2 of those columns.
    The ordering is determined by first ranking the data in row `m1`,
    then ranking any tied columns according to the data in row `m1`+1,
    and so on up to row m2.

    ``intmat_rank_columns`` uses a variant of list-merging, as described
    on pages 165--166 in Knuth (1973).
    The function takes advantage of natural ordering in the data, and
    uses a simple list insertion in a preparatory pass to generate
    ordered lists of length at least 10.
    The ranking is stable: equal columns preserve their ordering in the
    input data.

    References
    ----------
    Knuth, D E, 1973, `The Art of Computer Programming (Volume 3)`, (2nd
    Edition), Addison--Wesley
    """

    fun_name = 'naginterfaces.library.sort.intmat_rank_columns'

    _im = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'im'},
        im,
        copy=False,
    )
    _m2_py = _b_utils._EngineIntArrayType.get_shape(_im, {'fun_name': fun_name, 'entity': 'im'}, exp_rank=2)[0]
    _m2 = _b_utils._EngineIntScalarType(
        _m2_py, {'fun_name': fun_name, 'entity': '_m2'},
    ).to_ctypes()
    _n2_py = _b_utils._EngineIntArrayType.get_shape(_im, {'fun_name': fun_name, 'entity': 'im'}, exp_rank=2)[1]
    _n2 = _b_utils._EngineIntScalarType(
        _n2_py, {'fun_name': fun_name, 'entity': '_n2'},
    ).to_ctypes()
    _irank = _b_utils._EngineIntArrayType.empty_ndarray(
        (_n2_py,),
    )
    _b_sort.intmat_rank_columns(
        _im,
        m1,
        _m2,
        n1,
        _n2,
        order,
        _irank,
    )
    _irank_py = _irank
    return _irank_py

def arbitrary_rank(compar, m1, m2, data=None):
    u"""
    Rank arbitrary data.

    ``arbitrary_rank`` ranks arbitrary data according to a user-supplied
    comparison function.

    For full information please refer to the NAG Library document for
    m01dz

    https://www.nag.com/numeric/nl/nagdoc_29.2/flhtml/m01/m01dzf.html

    Parameters
    ----------
    compar : callable retval = compar(i, j, data=None)
        `compar` must specify the relative ordering of any two data
        items; it must return True if item `i` must come strictly
        **after** item `j` in the rank ordering.

        Parameters
        ~~~~~~~~~~
        i : int
            `i` and `j` identify the data items to be compared.

        j : int
            `i` and `j` identify the data items to be compared.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        retval : bool
            Must be True if item `i` must come strictly **after** item
            `j` in the rank ordering.

    m1 : int
        Note: this argument represents an array index; the value you
        supply must be base-1 for compatibility with the NAG Engine.

        `m1` must specify the lower bound of the range of data items to
        be ranked, and the range of ranks to be assigned. Specifically,
        ``arbitrary_rank`` ranks the data items identified by integers
        in the range `m1` to `m2`, and assigns ranks in the range `m1`
        to `m2` which are stored in elements `m1`-1 to `m2`-1 of
        `irank`.

    m2 : int
        Note: this argument represents an array index; the value you
        supply must be base-1 for compatibility with the NAG Engine.

        `m2` must specify the upper bound of the range of data items to
        be ranked, and the range of ranks to be assigned. Specifically,
        ``arbitrary_rank`` ranks the data items identified by integers
        in the range `m1` to `m2`, and assigns ranks in the range `m1`
        to `m2` which are stored in elements `m1`-1 to `m2`-1 of
        `irank`.

    data : arbitrary, optional
        User-communication data for callback functions.

    Returns
    -------
    irank : int, ndarray, shape (`m2`)
        Elements `m1`-1 to `m2`-1 of `irank` contain the ranks of the
        data items `m1` to `m2`. Note that the ranks are in the range
        `m1` to `m2`: thus, if item i is first in the rank ordering,
        `irank`[i-1] contains `m1`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `m1` = *<value>* and `m2` = *<value>*.

            Constraint: `m1` <= `m2`.

        (`errno` 1)
            On entry, `m1` = *<value>*.

            Constraint: `m1` >= 1.

        (`errno` 1)
            On entry, `m2` = *<value>*.

            Constraint: `m2` >= 1.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    ``arbitrary_rank`` is a general purpose function for ranking
    arbitrary data. ``arbitrary_rank`` does not access the data
    directly; instead it calls `compar` to determine the relative
    ordering of any two data items.
    The data items are identified simply by an integer in the range `m1`
    to `m2`.

    ``arbitrary_rank`` uses a variant of list-merging, as described on
    pages 165--166 in Knuth (1973).
    The function takes advantage of natural ordering in the data, and
    uses a simple list insertion in a preparatory pass to generate
    ordered lists of length at least 10.

    References
    ----------
    Knuth, D E, 1973, `The Art of Computer Programming (Volume 3)`, (2nd
    Edition), Addison--Wesley
    """

    def cb_compar(i, j, data=None):
        """Wrapping user's compar. Called as base compar."""
        if data is not None:
            _retval_compar = compar(
                i,
                j,
                data,
            )
        else:
            _retval_compar = compar(
                i,
                j,
            )
        return _retval_compar

    fun_name = 'naginterfaces.library.sort.arbitrary_rank'

    _b_utils._check_callable(compar, {'fun_name': fun_name, 'entity': 'compar'})
    _compar = cb_compar
    _m2 = _b_utils._EngineIntScalarType(
        m2, {'fun_name': fun_name, 'entity': 'm2'},
    ).to_ctypes()
    _m2_py = _b_utils._EngineIntScalarType.to_py(_m2)
    _irank = _b_utils._EngineIntArrayType.empty_ndarray(
        (_m2_py,),
    )
    _b_sort.arbitrary_rank(
        _compar,
        m1,
        _m2,
        _irank,
        data,
    )
    _irank_py = _irank
    return _irank_py

RealvecRankRearrangeReturnData = _collections.namedtuple(
    'RealvecRankRearrangeReturnData',
    ['rv', 'irank'],
)

def realvec_rank_rearrange(rv, m1, irank):
    u"""
    Rearrange a vector according to given ranks, real numbers.

    ``realvec_rank_rearrange`` rearranges a vector of real numbers into
    the order specified by a vector of ranks.

    For full information please refer to the NAG Library document for
    m01ea

    https://www.nag.com/numeric/nl/nagdoc_29.2/flhtml/m01/m01eaf.html

    Parameters
    ----------
    rv : float, array-like, shape (m2)
        Elements `m1`-1 to m2-1 of `rv` must contain real values to be
        rearranged.

    m1 : int
        Note: this argument represents an array index; the value you
        supply must be base-1 for compatibility with the NAG Engine.

        `m1` and m2 must specify the range of the ranks supplied in
        `irank` and the elements of `rv` to be rearranged.

    irank : int, array-like, shape (m2)
        Elements `m1`-1 to m2-1 of `irank` must contain a permutation of
        the integers `m1` to m2, which are interpreted as a vector of
        ranks.

    Returns
    -------
    rv : float, ndarray, shape (m2)
        These values are rearranged into rank order. For example, if
        `irank`[i-1] = `m1`, then the initial value of `rv`[i-1] is
        moved to `rv`[`m1`-1].

    irank : int, ndarray, shape (m2)
        Used as internal workspace prior to being restored and hence is
        unchanged.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `m1` = *<value>* and m2 = *<value>*.

            Constraint: `m1` <= m2.

        (`errno` 1)
            On entry, `m1` = *<value>*.

            Constraint: `m1` >= 1.

        (`errno` 1)
            On entry, m2 = *<value>*.

            Constraint: m2 >= 1.

        (`errno` 2)
            `irank`[`m1`:m2-1] does not contain a permutation of the
            integers `m1` to m2. `irank`[I-1] contains an out-of-range
            value: I = *<value>*, `irank`[I-1] = *<value>*.

        (`errno` 3)
            `irank`[`m1`:m2-1] does not contain a permutation of the
            integers `m1` to m2. `irank` contains a repeated value:
            *<value>*.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    ``realvec_rank_rearrange`` is designed to be used typically in
    conjunction with the ranking functions ``realvec_rank``,
    ``intvec_rank``, ``charvec_rank``, ``realmat_rank_rows``,
    ``intmat_rank_rows``, ``realmat_rank_columns``,
    ``intmat_rank_columns`` and ``arbitrary_rank``.
    After one of those functions has been called to determine a vector
    of ranks, ``realvec_rank_rearrange`` can be called to rearrange a
    vector of real numbers into the rank order.
    If the vector of ranks has been generated in some other way, then
    ``permute_check`` should be called to check its validity before
    ``realvec_rank_rearrange`` is called.
    """

    fun_name = 'naginterfaces.library.sort.realvec_rank_rearrange'

    _rv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'rv'},
        rv,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m2_py = _b_utils._EngineFloat64ArrayType.get_shape(_rv, {'fun_name': fun_name, 'entity': 'rv'}, exp_rank=1)[0]
    _m2 = _b_utils._EngineIntScalarType(
        _m2_py, {'fun_name': fun_name, 'entity': '_m2'},
    ).to_ctypes()
    _irank = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'irank'},
        irank,
        copy=True,
    )
    _b_sort.realvec_rank_rearrange(
        _rv,
        m1,
        _m2,
        _irank,
    )
    _rv_py = _rv
    _irank_py = _irank
    _return_data = RealvecRankRearrangeReturnData(
        _rv_py,
        _irank_py,
    )
    return _return_data

IntvecRankRearrangeReturnData = _collections.namedtuple(
    'IntvecRankRearrangeReturnData',
    ['iv', 'irank'],
)

def intvec_rank_rearrange(iv, m1, irank):
    u"""
    Rearrange a vector according to given ranks, integer numbers.

    ``intvec_rank_rearrange`` rearranges a vector of integer numbers
    into the order specified by a vector of ranks.

    For full information please refer to the NAG Library document for
    m01eb

    https://www.nag.com/numeric/nl/nagdoc_29.2/flhtml/m01/m01ebf.html

    Parameters
    ----------
    iv : int, array-like, shape (m2)
        Elements `m1`-1 to m2-1 of `iv` must contain integer values to
        be rearranged.

    m1 : int
        Note: this argument represents an array index; the value you
        supply must be base-1 for compatibility with the NAG Engine.

        `m1` and m2 specify the range of the ranks supplied in `irank`
        and the elements of `iv` to be rearranged.

    irank : int, array-like, shape (m2)
        Elements `m1`-1 to m2-1 of `irank` must contain a permutation of
        the integers `m1` to m2, which are interpreted as a vector of
        ranks.

    Returns
    -------
    iv : int, ndarray, shape (m2)
        These values are rearranged into rank order. For example, if
        `irank`[i-1] = `m1`, then the initial value of `iv`[i-1] is
        moved to `iv`[`m1`-1].

    irank : int, ndarray, shape (m2)
        Used as internal workspace prior to being restored and hence is
        unchanged.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `m1` = *<value>* and m2 = *<value>*.

            Constraint: `m1` <= m2.

        (`errno` 1)
            On entry, `m1` = *<value>*.

            Constraint: `m1` >= 1.

        (`errno` 1)
            On entry, m2 = *<value>*.

            Constraint: m2 >= 1.

        (`errno` 2)
            `irank`[`m1`:m2-1] does not contain a permutation of the
            integers `m1` to m2. `irank`[I-1] contains an out-of-range
            value: I = *<value>*, `irank`[I-1] = *<value>*.

        (`errno` 3)
            `irank`[`m1`:m2-1] does not contain a permutation of the
            integers `m1` to m2. `irank` contains a repeated value:
            *<value>*.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    ``intvec_rank_rearrange`` is designed to be used typically in
    conjunction with the ranking functions ``realvec_rank``,
    ``intvec_rank``, ``charvec_rank``, ``realmat_rank_rows``,
    ``intmat_rank_rows``, ``realmat_rank_columns``,
    ``intmat_rank_columns`` and ``arbitrary_rank``.
    After one of those functions has been called to determine a vector
    of ranks, ``intvec_rank_rearrange`` can be called to rearrange a
    vector of integer numbers into the rank order.
    If the vector of ranks has been generated in some other way, then
    ``permute_check`` should be called to check its validity before
    ``intvec_rank_rearrange`` is called.
    """

    fun_name = 'naginterfaces.library.sort.intvec_rank_rearrange'

    _iv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'iv'},
        iv,
        copy=True,
    )
    _m2_py = _b_utils._EngineIntArrayType.get_shape(_iv, {'fun_name': fun_name, 'entity': 'iv'}, exp_rank=1)[0]
    _m2 = _b_utils._EngineIntScalarType(
        _m2_py, {'fun_name': fun_name, 'entity': '_m2'},
    ).to_ctypes()
    _irank = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'irank'},
        irank,
        copy=True,
    )
    _b_sort.intvec_rank_rearrange(
        _iv,
        m1,
        _m2,
        _irank,
    )
    _iv_py = _iv
    _irank_py = _irank
    _return_data = IntvecRankRearrangeReturnData(
        _iv_py,
        _irank_py,
    )
    return _return_data

CharvecRankRearrangeReturnData = _collections.namedtuple(
    'CharvecRankRearrangeReturnData',
    ['ch', 'irank'],
)

def charvec_rank_rearrange(ch, m1, irank):
    u"""
    Rearrange a vector according to given ranks, character data.

    ``charvec_rank_rearrange`` rearranges a vector of character data
    into the order specified by a vector of ranks.

    For full information please refer to the NAG Library document for
    m01ec

    https://www.nag.com/numeric/nl/nagdoc_29.2/flhtml/m01/m01ecf.html

    Parameters
    ----------
    ch : str, array-like, shape (m2)
        Elements `m1`-1 to m2-1 of `ch` must contain character data to
        be rearranged.

    m1 : int
        Note: this argument represents an array index; the value you
        supply must be base-1 for compatibility with the NAG Engine.

        The range of the ranks supplied in `irank` and the elements of
        `ch` to be rearranged.

    irank : int, array-like, shape (m2)
        Elements `m1`-1 to m2-1 of `irank` must contain a permutation of
        the integers `m1` to m2, which are interpreted as a vector of
        ranks.

    Returns
    -------
    ch : str, ndarray, shape (m2)
        These values are rearranged into rank order. For example, if
        `irank`[i-1] = `m1`, then the initial value of `ch`[i-1] is
        moved to `ch`[`m1`-1].

    irank : int, ndarray, shape (m2)
        Used as internal workspace prior to being restored and hence is
        unchanged.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `m1` = *<value>* and m2 = *<value>*.

            Constraint: `m1` <= m2.

        (`errno` 1)
            On entry, `m1` = *<value>*.

            Constraint: `m1` >= 1.

        (`errno` 1)
            On entry, m2 = *<value>*.

            Constraint: m2 >= 1.

        (`errno` 2)
            On entry, len(`ch`[0]) = *<value>*.

            Constraint: len(`ch`[0]) <= 255.

        (`errno` 3)
            `irank`[`m1`:m2-1] does not contain a permutation of the
            integers `m1` to m2. `irank`[I-1] contains an out-of-range
            value: I = *<value>*, `irank`[I-1] = *<value>*.

        (`errno` 4)
            `irank`[`m1`:m2-1] does not contain a permutation of the
            integers `m1` to m2. `irank` contains a repeated value:
            *<value>*.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    ``charvec_rank_rearrange`` is designed to be used typically in
    conjunction with the ranking functions ``realvec_rank``,
    ``intvec_rank``, ``charvec_rank``, ``realmat_rank_rows``,
    ``intmat_rank_rows``, ``realmat_rank_columns``,
    ``intmat_rank_columns`` and ``arbitrary_rank``.
    After one of those functions has been called to determine a vector
    of ranks, ``charvec_rank_rearrange`` can be called to rearrange a
    vector of character data into the rank order.
    If the vector of ranks has been generated in some other way, then
    ``permute_check`` should be called to check its validity before
    ``charvec_rank_rearrange`` is called.
    """

    fun_name = 'naginterfaces.library.sort.charvec_rank_rearrange'

    _ch = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ch'},
        ch,
        copy=True,
    )
    _m2_py = _b_utils._EngineCharArrayType.get_shape(_ch, {'fun_name': fun_name, 'entity': 'ch'}, exp_rank=1)[0]
    _m2 = _b_utils._EngineIntScalarType(
        _m2_py, {'fun_name': fun_name, 'entity': '_m2'},
    ).to_ctypes()
    _irank = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'irank'},
        irank,
        copy=True,
    )
    _b_sort.charvec_rank_rearrange(
        _ch,
        m1,
        _m2,
        _irank,
    )
    _ch_py = _ch
    _irank_py = _irank
    _return_data = CharvecRankRearrangeReturnData(
        _ch_py,
        _irank_py,
    )
    return _return_data

CmplxvecRankRearrangeReturnData = _collections.namedtuple(
    'CmplxvecRankRearrangeReturnData',
    ['cv', 'irank'],
)

def cmplxvec_rank_rearrange(cv, m1, irank):
    u"""
    Rearrange a vector according to given ranks, complex numbers.

    ``cmplxvec_rank_rearrange`` rearranges a vector of complex numbers
    into the order specified by a vector of ranks.

    For full information please refer to the NAG Library document for
    m01ed

    https://www.nag.com/numeric/nl/nagdoc_29.2/flhtml/m01/m01edf.html

    Parameters
    ----------
    cv : complex, array-like, shape (m2)
        Elements `m1`-1 to m2-1 of `cv` must contain complex values to
        be rearranged.

    m1 : int
        Note: this argument represents an array index; the value you
        supply must be base-1 for compatibility with the NAG Engine.

        `m1` and m2 must specify the range of the ranks supplied in
        `irank` and the elements of `cv` to be rearranged.

    irank : int, array-like, shape (m2)
        Elements `m1`-1 to m2-1 of `irank` must contain a permutation of
        the integers `m1` to m2, which are interpreted as a vector of
        ranks.

    Returns
    -------
    cv : complex, ndarray, shape (m2)
        These values are rearranged into rank order. For example, if
        `irank`[i-1] = `m1`, then the initial value of `cv`[i-1] is
        moved to `cv`[`m1`-1].

    irank : int, ndarray, shape (m2)
        Used as internal workspace prior to being restored and hence is
        unchanged.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `m1` = *<value>* and m2 = *<value>*.

            Constraint: `m1` <= m2.

        (`errno` 1)
            On entry, `m1` = *<value>*.

            Constraint: `m1` >= 1.

        (`errno` 1)
            On entry, m2 = *<value>*.

            Constraint: m2 >= 1.

        (`errno` 2)
            `irank`[`m1`:m2-1] does not contain a permutation of the
            integers `m1` to m2. `irank`[I-1] contains an out-of-range
            value: I = *<value>*, `irank`[I-1] = *<value>*.

        (`errno` 3)
            `irank`[`m1`:m2-1] does not contain a permutation of the
            integers `m1` to m2. `irank` contains a repeated value:
            *<value>*.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    ``cmplxvec_rank_rearrange`` is designed to be used typically in
    conjunction with the ranking functions ``realvec_rank``,
    ``intvec_rank``, ``charvec_rank``, ``realmat_rank_rows``,
    ``intmat_rank_rows``, ``realmat_rank_columns``,
    ``intmat_rank_columns`` and ``arbitrary_rank``.
    After one of those functions has been called to determine a vector
    of ranks, ``cmplxvec_rank_rearrange`` can be called to rearrange a
    vector of complex numbers into the rank order.
    If the vector of ranks has been generated in some other way, then
    ``permute_check`` should be called to check its validity before
    ``cmplxvec_rank_rearrange`` is called.
    """

    fun_name = 'naginterfaces.library.sort.cmplxvec_rank_rearrange'

    _cv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'cv'},
        cv,
        copy=True,
    )
    _m2_py = _b_utils._EngineComplex128ArrayType.get_shape(_cv, {'fun_name': fun_name, 'entity': 'cv'}, exp_rank=1)[0]
    _m2 = _b_utils._EngineIntScalarType(
        _m2_py, {'fun_name': fun_name, 'entity': '_m2'},
    ).to_ctypes()
    _irank = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'irank'},
        irank,
        copy=True,
    )
    _b_sort.cmplxvec_rank_rearrange(
        _cv,
        m1,
        _m2,
        _irank,
    )
    _cv_py = _cv
    _irank_py = _irank
    _return_data = CmplxvecRankRearrangeReturnData(
        _cv_py,
        _irank_py,
    )
    return _return_data

def realvec_search(valid, rv, item, m1=1):
    u"""
    Binary search in set of real numbers.

    ``realvec_search`` searches an ordered vector of `float` numbers and
    returns the index of the first value equal to the sought-after item.

    For full information please refer to the NAG Library document for
    m01na

    https://www.nag.com/numeric/nl/nagdoc_29.2/flhtml/m01/m01naf.html

    Parameters
    ----------
    valid : bool
        If `valid` is set to True argument checking will be performed.
        If `valid` is set to False ``realvec_search`` will be called
        without argument checking (which includes checking that array
        `rv` is sorted in ascending order) and the function will return
        with no exception or warning is raised. See [Further Comments]
        for further details.

    rv : float, array-like, shape (m2)
        Elements `m1`-1 to m2-1 contain `float` values to be searched.

    item : float
        The sought-after item.

    m1 : int, optional
        Note: this argument represents an array index; the value you
        supply must be base-1 for compatibility with the NAG Engine.

        The index of the first element of `rv` to be searched.

    Returns
    -------
    irv : int
        Note: this argument represents an array index; the value
        returned will be base-1.

        The index of the first value equal to the sought-after item.

    Raises
    ------
    NagValueError
        (`errno` 2)
            On entry, `rv` must be sorted in ascending order: `rv`
            element *<value>* > element *<value>*.

        (`errno` 3)
            On entry, `m1` = *<value>*.

            Constraint: `m1` >= 1.

        (`errno` 4)
            On entry, `m1` = *<value>*, m2 = *<value>*.

            Constraint: `m1` <= m2.

    Notes
    -----
    ``realvec_search`` is based on Professor Niklaus Wirth's
    implementation of the Binary Search algorithm (see Wirth (2004)),
    but with two modifications.
    First, if the sought-after item is less than the value of the first
    element of the array to be searched, -1 is returned.
    Second, if a value equal to the sought-after item is not found, the
    index of the immediate lower value is returned.

    References
    ----------
    Wirth, N, 2004, `Algorithms and Data Structures`, 35--36, Prentice
    Hall
    """

    fun_name = 'naginterfaces.library.sort.realvec_search'

    _rv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'rv'},
        rv,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m2_py = _b_utils._EngineFloat64ArrayType.get_shape(_rv, {'fun_name': fun_name, 'entity': 'rv'}, exp_rank=1)[0]
    _m2 = _b_utils._EngineIntScalarType(
        _m2_py, {'fun_name': fun_name, 'entity': '_m2'},
    ).to_ctypes()
    _return_scalar = _b_sort.realvec_search(
        valid,
        _rv,
        m1,
        _m2,
        item,
    )
    return _return_scalar

def intvec_search(valid, iv, item, m1=1):
    u"""
    Binary search in set of integer numbers.

    ``intvec_search`` searches an ordered vector of integer numbers and
    returns the index of the first value equal to the sought-after item.

    For full information please refer to the NAG Library document for
    m01nb

    https://www.nag.com/numeric/nl/nagdoc_29.2/flhtml/m01/m01nbf.html

    Parameters
    ----------
    valid : bool
        If `valid` is set to True argument checking will be performed.
        If `valid` is set to False ``intvec_search`` will be called
        without argument checking (which includes checking that array
        `iv` is sorted in ascending order) and the function will return
        with no exception or warning is raised. See [Further Comments]
        for further details.

    iv : int, array-like, shape (m2)
        Elements `m1`-1 to m2-1 contain integer values to be searched.

    item : int
        The sought-after item.

    m1 : int, optional
        Note: this argument represents an array index; the value you
        supply must be base-1 for compatibility with the NAG Engine.

        The index of the first element of `iv` to be searched.

    Returns
    -------
    iiv : int
        Note: this argument represents an array index; the value
        returned will be base-1.

        The index of the first value equal to the sought-after item.

    Raises
    ------
    NagValueError
        (`errno` 2)
            On entry, `iv` must be sorted in ascending order: `iv`
            element *<value>* > element *<value>*.

        (`errno` 3)
            On entry, `m1` = *<value>*.

            Constraint: `m1` >= 1.

        (`errno` 4)
            On entry, `m1` = *<value>*, m2 = *<value>*.

            Constraint: `m1` <= m2.

    Notes
    -----
    ``intvec_search`` is based on Professor Niklaus Wirth's
    implementation of the Binary Search algorithm (see Wirth (2004)),
    but with two modifications.
    First, if the sought-after item is less than the value of the first
    element of the array to be searched, -1 is returned.
    Second, if a value equal to the sought-after item is not found, the
    index of the immediate lower value is returned.

    References
    ----------
    Wirth, N, 2004, `Algorithms and Data Structures`, 35--36, Prentice
    Hall
    """

    fun_name = 'naginterfaces.library.sort.intvec_search'

    _iv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'iv'},
        iv,
        copy=False,
    )
    _m2_py = _b_utils._EngineIntArrayType.get_shape(_iv, {'fun_name': fun_name, 'entity': 'iv'}, exp_rank=1)[0]
    _m2 = _b_utils._EngineIntScalarType(
        _m2_py, {'fun_name': fun_name, 'entity': '_m2'},
    ).to_ctypes()
    _return_scalar = _b_sort.intvec_search(
        valid,
        _iv,
        m1,
        _m2,
        item,
    )
    return _return_scalar

def charvec_search(valid, ch, item, m1=1):
    u"""
    Binary search in set of character data.

    ``charvec_search`` examines an ordered vector of strings and returns
    the index of the first value equal to the sought-after item.
    Character items are compared according to the ASCII collating
    sequence.

    For full information please refer to the NAG Library document for
    m01nc

    https://www.nag.com/numeric/nl/nagdoc_29.2/flhtml/m01/m01ncf.html

    Parameters
    ----------
    valid : bool
        If `valid` is set to True argument checking will be performed.
        If `valid` is set to False ``charvec_search`` will be called
        without argument checking, which includes checking that array
        `ch` is sorted in ascending order and the function will return
        with no exception or warning is raised. See [Further Comments]
        for further details.

    ch : str, array-like, shape (m2)
        Elements `m1`-1 to m2-1 contain character data to be searched.

    item : str
        The sought-after item. Trailing space characters are ignored.

    m1 : int, optional
        Note: this argument represents an array index; the value you
        supply must be base-1 for compatibility with the NAG Engine.

        The index of the first element of `ch` to be searched.

    Returns
    -------
    ich : int
        Note: this argument represents an array index; the value
        returned will be base-1.

        The index of the first value equal to the sought-after item.

    Raises
    ------
    NagValueError
        (`errno` 2)
            On entry, `ch` must be sorted in ascending order: `ch`
            element *<value>* > element *<value>*.

        (`errno` 3)
            On entry, `m1` = *<value>*.

            Constraint: `m1` >= 1.

        (`errno` 4)
            On entry, `m1` = *<value>*, m2 = *<value>*.

            Constraint: m2 >= `m1`.

        (`errno` 5)
            On entry, the length of each element of `ch` must be at most
            255: maximum string length = *<value>*.

    Notes
    -----
    ``charvec_search`` is based on Professor Niklaus Wirth's
    implementation of the Binary Search algorithm (see Wirth (2004)),
    but with two modifications.
    First, if the sought-after item is less than the value of the first
    element of the array to be searched, -1 is returned.
    Second, if a value equal to the sought-after item is not found, the
    index of the immediate lower value is returned.

    References
    ----------
    Wirth, N, 2004, `Algorithms and Data Structures`, 35--36, Prentice
    Hall
    """

    fun_name = 'naginterfaces.library.sort.charvec_search'

    _ch = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ch'},
        ch,
        copy=False,
    )
    _m2_py = _b_utils._EngineCharArrayType.get_shape(_ch, {'fun_name': fun_name, 'entity': 'ch'}, exp_rank=1)[0]
    _m2 = _b_utils._EngineIntScalarType(
        _m2_py, {'fun_name': fun_name, 'entity': '_m2'},
    ).to_ctypes()
    _return_scalar = _b_sort.charvec_search(
        valid,
        _ch,
        m1,
        _m2,
        item,
    )
    return _return_scalar

def realvec_vec_search(valid, rv, m1, m2, item, comm=None):
    u"""
    Searches an ordered set of real numbers using an O(1) method.

    ``realvec_vec_search`` searches a strictly ordered vector of `float`
    numbers and returns the indices of the largest numbers in the vector
    which are smaller than or equal to the sought-after items.

    For full information please refer to the NAG Library document for
    m01nd

    https://www.nag.com/numeric/nl/nagdoc_29.2/flhtml/m01/m01ndf.html

    Parameters
    ----------
    valid : bool
        If `valid` is set to True argument checking will be performed.
        If `valid` is set to False ``realvec_vec_search`` will be called
        without argument checking (which includes checking that array
        `rv` is sorted in strictly ascending order). See [Further
        Comments] for further details.

    rv : float, array-like, shape (n)
        X, the `float` values to be searched.

    m1 : int
        Note: this argument represents an array index; the value you
        supply must be base-1 for compatibility with the NAG Engine.

        The index of the first element of `rv` to be searched.

    m2 : int
        Note: this argument represents an array index; the value you
        supply must be base-1 for compatibility with the NAG Engine.

        The index of the last element of `rv` to be searched.

    item : float, array-like, shape (m)
        Z, the sought-after items.

    comm : None or dict, communication object, optional, modified in
    place
        Communication structure.

        If `comm` is None then searching is performed using an
        offset-based binary search.

        Otherwise, `comm` may be supplied uninitialized.

        It will then be populated and the routine will perform a direct
        search of `rv`.

        The populated `comm` can then be supplied again to subsequent
        calls for repeated searching of `rv`.

    Returns
    -------
    idx : int, ndarray, shape (m)
        Note: this argument represents array indices; the values
        returned will be base-1.

        `idx`[i-1] contains the index of the largest entry in `rv` which
        is smaller than or equal to `item`[i-1].

    Raises
    ------
    NagValueError
        (`errno` 2)
            On entry, `rv` must be sorted in strictly ascending order:
            `rv` element *<value>* > element *<value>*.

        (`errno` 3)
            On entry, `m1` = *<value>*.

            Constraint: `m1` >= 0.

        (`errno` 4)
            On entry, `m1` = *<value>*, `m2` = *<value>*, n = *<value>*.

            Constraint: `m1` <= `m2` <= n.

        (`errno` 5)
            On entry, n = *<value>*.

            Constraint: n >= 2.

        (`errno` 6)
            On entry, m = *<value>*.

            Constraint: m >= 1.

        (`errno` 7)
            On entry, the values in `rv` are such that the required
            value of lk would overflow the current platform's largest
            positive integer value.

        (`errno` 8)
            `comm` may have become corrupted.

    Notes
    -----
    ``realvec_vec_search`` searches an array, X, of n
    strictly-increasing `float` numbers, X_i < X_{i+1}, i = 1...(n-1),
    for the elements of an unordered array of m `float` numbers, Z.
    If a value equal to a sought-after item Z_i is not found in X, the
    index of the immediate lower value is returned.
    If Z_i is less than X_1, -1 is returned.

    The function implements the direct search algorithm presented as
    Algorithm 8 in Cannizzo (2018).
    It precomputes a scalar, H, and a reference vector of indices, K,
    that it uses to speed up searches of X.
    The length of K is a function of the number of entries in X and
    their values, and the function provides a mechanism to compute what
    this length should be.

    If the amount of memory required for K is infeasible,
    ``realvec_vec_search`` implements offset-based binary search
    (Algorithm 5 in Cannizzo (2018)) as an alternative that does not
    require K to be used.

    See [Further Comments] for more information on the relative time
    complexities of the two search algorithms.

    References
    ----------
    Cannizzo, F, 2018, `A fast and vectorizable alternative to binary
    search in O(1) with wide applicability to arrays of floating point
    numbers`, Journal of Parallel and Distributed Computing (113),
    37--54
    """

    fun_name = 'naginterfaces.library.sort.realvec_vec_search'

    _rv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'rv'},
        rv,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_rv, {'fun_name': fun_name, 'entity': 'rv'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _item = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'item'},
        item,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_py = _b_utils._EngineFloat64ArrayType.get_shape(_item, {'fun_name': fun_name, 'entity': 'item'}, exp_rank=1)[0]
    _m = _b_utils._EngineIntScalarType(
        _m_py, {'fun_name': fun_name, 'entity': '_m'},
    ).to_ctypes()
    _idx = _b_utils._EngineIntArrayType.empty_ndarray(
        (_m_py,),
    )
    _b_sort.realvec_vec_search(
        valid,
        _rv,
        _n,
        m1,
        m2,
        _item,
        _m,
        _idx,
        comm,
    )
    _idx_py = _idx
    return _idx_py

def permute_invert(iperm, m1):
    u"""
    Invert a permutation.

    ``permute_invert`` inverts a permutation, and hence converts a rank
    vector to an index vector, or vice versa.

    For full information please refer to the NAG Library document for
    m01za

    https://www.nag.com/numeric/nl/nagdoc_29.2/flhtml/m01/m01zaf.html

    Parameters
    ----------
    iperm : int, array-like, shape (m2)
        Elements `m1`-1 to m2-1 of `iperm` must contain a permutation of
        the integers `m1` to m2.

    m1 : int
        Note: this argument represents an array index; the value you
        supply must be base-1 for compatibility with the NAG Engine.

        `m1` and m2 must specify the range of elements used in the array
        `iperm` and the range of values in the permutation, as specified
        under `iperm`.

    Returns
    -------
    iperm : int, ndarray, shape (m2)
        These elements contain the inverse permutation of the integers
        `m1` to m2.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `m1` = *<value>* and m2 = *<value>*.

            Constraint: `m1` <= m2.

        (`errno` 1)
            On entry, `m1` = *<value>*.

            Constraint: `m1` >= 1.

        (`errno` 1)
            On entry, m2 = *<value>*.

            Constraint: m2 >= 1.

        (`errno` 2)
            `iperm`[`m1`:m2-1] does not contain a permutation of the
            integers `m1` to m2. `iperm`[I-1] contains an out-of-range
            value: I = *<value>*, `iperm`[I-1] = *<value>*.

        (`errno` 3)
            `iperm`[`m1`:m2-1] does not contain a permutation of the
            integers `m1` to m2. `iperm` contains a repeated value:
            *<value>*.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    There are two common ways of describing a permutation using an
    integer vector `iperm`.
    The first uses ranks: `iperm`[i-1] holds the position to which the
    ith data element should be moved in order to sort the data; in other
    words its rank in the sorted order.
    The second uses indices: `iperm`[i-1] holds the current position of
    the data element which would occur in ith position in sorted order.
    For example, given the values

        3.5 5.9 2.9 0.5

    to be sorted in ascending order, the ranks would be

        3 4 2 1

        2 3 1 0

    and the indices would be

        4 3 1 2

        3 2 0 1

    The functions ``realvec_rank``, ``intvec_rank``, ``charvec_rank``,
    ``realmat_rank_rows``, ``intmat_rank_rows``,
    ``realmat_rank_columns``, ``intmat_rank_columns`` and
    ``arbitrary_rank`` generate ranks, and the functions
    ``realvec_rank_rearrange``, ``intvec_rank_rearrange``,
    ``charvec_rank_rearrange`` and ``cmplxvec_rank_rearrange`` require
    ranks to be supplied to specify the reordering.
    However if it is desired simply to refer to the data in sorted order
    without actually reordering them, indices are more convenient than
    ranks.

    ``permute_invert`` can be used to convert ranks to indices, or
    indices to ranks, as the two permutations are inverses of each
    another.
    """

    fun_name = 'naginterfaces.library.sort.permute_invert'

    _iperm = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'iperm'},
        iperm,
        copy=True,
    )
    _m2_py = _b_utils._EngineIntArrayType.get_shape(_iperm, {'fun_name': fun_name, 'entity': 'iperm'}, exp_rank=1)[0]
    _m2 = _b_utils._EngineIntScalarType(
        _m2_py, {'fun_name': fun_name, 'entity': '_m2'},
    ).to_ctypes()
    _b_sort.permute_invert(
        _iperm,
        m1,
        _m2,
    )
    _iperm_py = _iperm
    return _iperm_py

def permute_check(iperm, m1):
    u"""
    Check validity of a permutation.

    ``permute_check`` checks the validity of a permutation.

    For full information please refer to the NAG Library document for
    m01zb

    https://www.nag.com/numeric/nl/nagdoc_29.2/flhtml/m01/m01zbf.html

    Parameters
    ----------
    iperm : int, array-like, shape (m2)
        Elements `m1`-1 to m2-1 of `iperm` must be set to values which
        are supposed to be a permutation of the integers `m1` to m2. If
        they are a valid permutation, the function exits with no
        exception or warning is raised.

    m1 : int
        Note: this argument represents an array index; the value you
        supply must be base-1 for compatibility with the NAG Engine.

        The range of elements used in the array `iperm` and the range of
        values in the permutation, as specified under `iperm`.

    Returns
    -------
    iperm : int, ndarray, shape (m2)
        Used as internal workpsace prior to being restored and hence is
        unchanged.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `m1` = *<value>* and m2 = *<value>*.

            Constraint: `m1` <= m2.

        (`errno` 1)
            On entry, `m1` = *<value>*.

            Constraint: `m1` >= 1.

        (`errno` 1)
            On entry, m2 = *<value>*.

            Constraint: m2 >= 1.

        (`errno` 2)
            `iperm`[`m1`:m2-1] does not contain a permutation of the
            integers `m1` to m2. `iperm`[I-1] contains an out-of-range
            value: I = *<value>*, `iperm`[I-1] = *<value>*.

        (`errno` 3)
            `iperm`[`m1`:m2-1] does not contain a permutation of the
            integers `m1` to m2. `iperm` contains a repeated value:
            *<value>*.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    ``permute_check`` can be used to check the validity of user-supplied
    ranks or indices, without the ranks or indices being corrupted.
    """

    fun_name = 'naginterfaces.library.sort.permute_check'

    _iperm = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'iperm'},
        iperm,
        copy=True,
    )
    _m2_py = _b_utils._EngineIntArrayType.get_shape(_iperm, {'fun_name': fun_name, 'entity': 'iperm'}, exp_rank=1)[0]
    _m2 = _b_utils._EngineIntScalarType(
        _m2_py, {'fun_name': fun_name, 'entity': '_m2'},
    ).to_ctypes()
    _b_sort.permute_check(
        _iperm,
        m1,
        _m2,
    )
    _iperm_py = _iperm
    return _iperm_py

PermuteDecomposeReturnData = _collections.namedtuple(
    'PermuteDecomposeReturnData',
    ['iperm', 'icycl'],
)

def permute_decompose(iperm, m1):
    u"""
    Decompose a permutation into cycles.

    ``permute_decompose`` decomposes a permutation into cycles, as an
    aid to reordering ranked data.

    For full information please refer to the NAG Library document for
    m01zc

    https://www.nag.com/numeric/nl/nagdoc_29.2/flhtml/m01/m01zcf.html

    Parameters
    ----------
    iperm : int, array-like, shape (m2)
        Elements `m1`-1 to m2-1 of `iperm` must contain a permutation of
        the integers `m1` to m2.

    m1 : int
        Note: this argument represents an array index; the value you
        supply must be base-1 for compatibility with the NAG Engine.

        `m1` and m2 must specify the range of elements used in the array
        `iperm` and the range of values in the permutation, as specified
        under `iperm`.

    Returns
    -------
    iperm : int, ndarray, shape (m2)
        Is used as internal workpsace prior to being restored and hence
        is unchanged.

    icycl : int, ndarray, shape (m2)
        Elements `m1`-1 to m2-1 of `icycl` contain a representation of
        the permutation as a list of cycles, with the first integer in
        each cycle negated. (See [Notes].)

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `m1` = *<value>* and m2 = *<value>*.

            Constraint: `m1` <= m2.

        (`errno` 1)
            On entry, `m1` = *<value>*.

            Constraint: `m1` >= 1.

        (`errno` 1)
            On entry, m2 = *<value>*.

            Constraint: m2 >= 1.

        (`errno` 2)
            `iperm`[`m1`:m2-1] does not contain a permutation of the
            integers `m1` to m2. `iperm`[I-1] contains an out-of-range
            value: I = *<value>*, `iperm`[I-1] = *<value>*.

        (`errno` 3)
            `iperm`[`m1`:m2-1] does not contain a permutation of the
            integers `m1` to m2. `iperm` contains a repeated value:
            *<value>*.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    ``permute_decompose`` is provided as an aid to reordering arbitrary
    data structures without using additional storage.
    However, you should consider carefully whether it is necessary to
    rearrange yourr data, or whether it would be simpler and more
    efficient to refer to the data in sorted order using an index
    vector, or to create a copy of the data in sorted order.

    To rearrange data into a different order without using additional
    storage, the simplest method is to decompose the permutation which
    specifies the new order into cycles and then to do a cyclic
    permutation of the data items in each cycle. (This is the method
    used by the reordering functions ``realvec_rank_rearrange``,
    ``intvec_rank_rearrange``, ``charvec_rank_rearrange`` and
    ``cmplxvec_rank_rearrange``.) Given a vector IRANK which specifies
    the ranks of the data (as generated by the functions
    ``realvec_rank``, ``intvec_rank``, ``charvec_rank``,
    ``realmat_rank_rows``, ``intmat_rank_rows``,
    ``realmat_rank_columns``, ``intmat_rank_columns`` and
    ``arbitrary_rank``), ``permute_decompose`` generates a new vector
    `icycl`, in which the permutation is represented in its component
    cycles, with the first element of each cycle negated.
    For example, the permutation

        5 7 4 2 1 6 3

    is composed of the cycles

        (1 5) (2 7 3 4) (6)

    and the vector `icycl` generated by ``permute_decompose`` contains

        -1 5 -2 7 3 4 -6

    In order to rearrange the data according to the specified ranks:

        item 6 must be left in place;

        items 1 and 5 must be interchanged;

        items 4, 2, 7 and 3 must be moved right one place round the
        cycle.

    The complete rearrangement can be achieved by the following logic:

        loop for k = `m1` to m2

            set i = `icycl`[k-1]

            if i < 0 then

                set j = -i

            else

                swap items i and j
    """

    fun_name = 'naginterfaces.library.sort.permute_decompose'

    _iperm = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'iperm'},
        iperm,
        copy=True,
    )
    _m2_py = _b_utils._EngineIntArrayType.get_shape(_iperm, {'fun_name': fun_name, 'entity': 'iperm'}, exp_rank=1)[0]
    _m2 = _b_utils._EngineIntScalarType(
        _m2_py, {'fun_name': fun_name, 'entity': '_m2'},
    ).to_ctypes()
    _icycl = _b_utils._EngineIntArrayType.empty_ndarray(
        (_m2_py,),
    )
    _b_sort.permute_decompose(
        _iperm,
        m1,
        _m2,
        _icycl,
    )
    _iperm_py = _iperm
    _icycl_py = _icycl
    _return_data = PermuteDecomposeReturnData(
        _iperm_py,
        _icycl_py,
    )
    return _return_data
